#include <sys/types.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>

#include "GPIOAccess.h"

#define REGISTER_BLOCK_ADDR     0x18040000
#define REGISTER_BLOCK_SIZE	0x30

#define REGISTER_OE_OFFSET	0
#define REGISTER_IN_OFFSET	1
#define REGISTER_OUT_OFFSET	2
#define REGISTER_SET_OFFSET	3
#define REGISTER_CLEAR_OFFSET	4

GPIOPwmPin * GPIOAccess::pwmPin[numPins];
bool GPIOAccess::pwmSetup = false;
bool GPIOAccess::pinUsable[numPins] = {
    true, // 0
    true, // 1
    false, // 2
    false, // 3
    false, // 4
    false, // 5
    true, // 6
    true, // 7
    true, // 8
    false, // 9
    false, // 10
    false, // 11
    true, // 12
    true, // 13
    true, // 14
    true, // 15
    true, // 16
    true, // 17
    true, // 18
    true, // 19
    false, // 20
    false, // 21
    false, // 22
    true, // 23
    false, // 24
    false, // 25
    true, // 26
};

volatile unsigned long int *GPIOAccess::regAddress = NULL;

GPIO_Result GPIOAccess::checkPinAndAccess(int pin) {
    if (!isPinUsable(pin)) {
        return GPIO_INVALID_PIN;
    }
    
    return checkAndSetupAddress(REGISTER_BLOCK_ADDR, REGISTER_BLOCK_SIZE);
}

GPIO_Result GPIOAccess::checkAndSetupAddress(unsigned long int blockBaseAddr, unsigned long int blockSize)
{
    if (regAddress != NULL) {
        if (regAddress == MAP_FAILED)
        {
                return GPIO_BAD_ACCESS;	// maybe return -2
        }

        return GPIO_OK;	// regAddress is now populated
    }
    
    int  m_mfd;

    if ((m_mfd = open("/dev/mem", O_RDWR)) < 0)
    {
            return GPIO_BAD_ACCESS;	// maybe return -1
    }

    regAddress = (unsigned long*)mmap(NULL, 
                                    blockSize, 
                                    PROT_READ|PROT_WRITE, 
                                    MAP_SHARED, 
                                    m_mfd, 
                                    blockBaseAddr
                            );
    close(m_mfd);

    if (regAddress == MAP_FAILED)
    {
            return GPIO_BAD_ACCESS;	// maybe return -2
    }

    if (!pwmSetup) {
        int i;
        for (i = 0; i < numPins; i++) {
            if (isPinUsable(i)) {
                pwmPin[i] = new GPIOPwmPin(i);
            } else {
                pwmPin[i] = NULL;
            }
        }
        pwmSetup = true;
        
    }
    
    return GPIO_OK;	// regAddress is now populated
}

void GPIOAccess::writeReg(unsigned long int registerOffset, unsigned long int value)
{
    *(regAddress + registerOffset) = value;
}

unsigned long int GPIOAccess::readReg(unsigned long int registerOffset)
{
    unsigned long int 	value = 0x0;

    // read the value 
    value = *(regAddress + registerOffset);

    return(value);
}

// change the value of a single bit
void GPIOAccess::setBit(unsigned long int &regVal, int bitNum, int value)
{
	if (value == 1) {
		regVal |= (1 << bitNum);
	}
	else {
		regVal &= ~(1 << bitNum);
	}

	// try this out
	// regVal ^= (-value ^ regVal) & (1 << bitNum);
}

// find the value of a single bit
int GPIOAccess::getBit(unsigned long int regVal, int bitNum)
{
	int value;

	// isolate the specific bit
	value = ((regVal >> bitNum) & 0x1);

	return (value);
}

GPIO_Result GPIOAccess::setDirection(int pinNum, GPIO_Direction dir) {
    GPIO_Result res = checkPinAndAccess(pinNum);
    if (res == GPIO_OK) {
        unsigned long int regVal;

        // read the current input and output settings
        regVal = readReg(REGISTER_OE_OFFSET);

        // set the OE for this pin
        setBit(regVal, pinNum, dir);

        // write the new register value
        writeReg(REGISTER_OE_OFFSET, regVal);
    }
    
    return res;
}

GPIO_Result GPIOAccess::getDirection(int pinNum, GPIO_Direction &dir) {
    GPIO_Result res = checkPinAndAccess(pinNum);
    if (res == GPIO_OK) {
        unsigned long int regVal;

        // read the current input and output settings
        regVal 	= readReg(REGISTER_OE_OFFSET);

        // find the OE for this pin
        dir = getDirectionImpl(pinNum);
    }
    
    return res;
}

GPIO_Direction GPIOAccess::getDirectionImpl(int pinNum) {
    unsigned long int regVal;

    // read the current input and output settings
    regVal = readReg(REGISTER_OE_OFFSET);

    // find the OE for this pin
    return getBit(regVal, pinNum) ? GPIO_OUTPUT : GPIO_INPUT;
}

GPIO_Result GPIOAccess::set(int pinNum, int value) {
    GPIO_Result res = checkPinAndAccess(pinNum);
    if (res == GPIO_OK) {
        if (getDirectionImpl(pinNum) != GPIO_OUTPUT) {
            return GPIO_INVALID_OP;
        }
        
        unsigned long int 	regAddr;
        unsigned long int 	regVal;

        if (value == 0 )	{
                // write to the clear register
                regAddr 	= REGISTER_CLEAR_OFFSET;
        }
        else {
                // write to the set register
                regAddr 	= REGISTER_SET_OFFSET;
        }

        // put the desired pin value into the register 
        regVal = (0x1 << pinNum);

        // write to the register
        writeReg (regAddr, regVal);
    }
    
    return res;
}

GPIO_Result GPIOAccess::get(int pinNum, int &value) {
    GPIO_Result res = checkPinAndAccess(pinNum);
    if (res == GPIO_OK) {
        if (getDirectionImpl(pinNum) != GPIO_INPUT) {
            return GPIO_INVALID_OP;
        }
        
        unsigned long int 	regVal;

        // read the current value of all pins
        regVal 	= readReg (REGISTER_IN_OFFSET);

        // find the value of the specified pin
        value 	= getBit(regVal, pinNum);
    }
    
    return res;
}

bool GPIOAccess::isPinUsable(int pin) {
    if ((pin < 0) || (pin >= numPins)) {
        return false;
    }
    
    return pinUsable[pin];
}

bool GPIOAccess::isAccessOk() {
    return checkAndSetupAddress(REGISTER_BLOCK_ADDR, REGISTER_BLOCK_SIZE) == GPIO_OK;
}

GPIO_Result GPIOAccess::setPWM(int pinNum, int freq, int duty) {
    GPIO_Result res = checkPinAndAccess(pinNum);
    if (res == GPIO_OK) {
        if (getDirectionImpl(pinNum) != GPIO_OUTPUT) {
            return GPIO_INVALID_OP;
        }
        
        res = pwmPin[pinNum]->setPWM(freq, duty);
        
        // Wait till thread has actually started PWM
        while (!pwmPin[pinNum]->isPWMRunning()) {
        }

        return res;
    }

    return res;
}

GPIO_Result GPIOAccess::startPWM(int pinNum) {
    GPIO_Result res = checkPinAndAccess(pinNum);
    if (res == GPIO_OK) {
        if (getDirectionImpl(pinNum) != GPIO_OUTPUT) {
            return GPIO_INVALID_OP;
        }

        return pwmPin[pinNum]->startPWM();
    }
    return res;
    
}

GPIO_Result GPIOAccess::stopPWM(int pinNum) {
    GPIO_Result res = checkPinAndAccess(pinNum);
    if (res == GPIO_OK) {
        if (getDirectionImpl(pinNum) != GPIO_OUTPUT) {
            return GPIO_INVALID_OP;
        }

        return pwmPin[pinNum]->stopPWM();
    }
    return res;
}

int GPIOAccess::getPWMFreq(int pinNum) {
    GPIO_Result res = checkPinAndAccess(pinNum);
    if (res == GPIO_OK) {
        if (getDirectionImpl(pinNum) != GPIO_OUTPUT) {
            return 0;
        }

        return pwmPin[pinNum]->getPWMFreq();
    }
    return 0;
}

int GPIOAccess::getPWMDuty(int pinNum) {
    GPIO_Result res = checkPinAndAccess(pinNum);
    if (res == GPIO_OK) {
        if (getDirectionImpl(pinNum) != GPIO_OUTPUT) {
            return 0;
        }

        return pwmPin[pinNum]->getPWMDuty();
    }
    return 0;
}

bool GPIOAccess::isPWMRunning(int pinNum) {
    GPIO_Result res = checkPinAndAccess(pinNum);
    if (res == GPIO_OK) {
        if (getDirectionImpl(pinNum) != GPIO_OUTPUT) {
            return false;
        }

        return pwmPin[pinNum]->isPWMRunning();
    }
    return false;
}
